@@ -1,75 +1,34 @@
 # -*- coding: utf-8 -*-
-"""Utilities for working with syntax."""
+"""Utilities for working with syntax.
+
+This module also contains the definitions for working with "decorated lambdas".
+"""
+
+__all__ = ["isec", "detect_callec",
+           "detect_lambda",
+           "is_decorator",
+           "is_lambda_decorator", "is_decorated_lambda", "destructure_decorated_lambda",
+           "has_tco", "has_curry", "has_deco",
+           "sort_lambda_decorators", "suggest_decorator_index",
+           "eliminate_ifones", "transform_statements",
+           "UnpythonicASTMarker", "UnpythonicExpandedMacroMarker",
+           "ExpandedContinuationsMarker", "ExpandedAutorefMarker"]
 
 from functools import partial
-import re
 
-from ast import Call, Name, Attribute, Lambda, FunctionDef, \
-                If, Num, NameConstant, For, While, With, Try, ClassDef, \
-                withitem
-from .astcompat import AsyncFunctionDef, AsyncFor, AsyncWith
+from ast import Call, Lambda, FunctionDef, AsyncFunctionDef, If, stmt
 
-from macropy.core import Captured
-from macropy.core.walkers import Walker
+from mcpyrate.core import add_postprocessor
+from mcpyrate.markers import ASTMarker, delete_markers
+from mcpyrate.quotes import is_captured_value
+from mcpyrate.walkers import ASTTransformer, ASTVisitor
 
+from .astcompat import getconstant
 from .letdoutil import isdo, ExpandedDoView
+from .nameutil import isx, getname
 
 from ..regutil import all_decorators, tco_decorators, decorator_registry
 
-def isx(tree, x, accept_attr=True):
-    """Test whether tree is a reference to the name ``x`` (str).
-
-    Alternatively, ``x`` may be a predicate that accepts a ``str``
-    and returns whether it matches, to support more complex matching
-    (e.g. ``lambda s: s.startswith("foo")``).
-
-    Both bare names and attributes can be recognized, to support
-    both from-imports and regular imports of ``somemodule.x``.
-
-    We support:
-
-        - bare name ``x``
-
-        - ``x`` inside a ``macropy.core.Captured`` node, which may be inserted
-          during macro expansion
-
-        - ``x`` as an attribute (if ``accept_attr=True``)
-    """
-    # WTF, so sometimes there **is** a Captured node, while sometimes there isn't (letdoutil.islet)? At which point are these removed?
-    # Captured nodes only come from unpythonic.syntax, and we use from-imports
-    # and bare names for anything hq[]'d; but any references that appear
-    # explicitly in the user code may use either bare names or somemodule.f.
-    ismatch = x if callable(x) else lambda s: s == x
-    return (type(tree) is Name and ismatch(tree.id)) or \
-           (type(tree) is Captured and ismatch(tree.name)) or \
-           (accept_attr and type(tree) is Attribute and ismatch(tree.attr))
-
-def make_isxpred(x):
-    """Make a predicate for isx.
-
-    Here ``x`` is an ``str``; the resulting function will match the names
-    x, x1, x2, ..., so that it works also with captured identifiers renamed
-    by MacroPy's ``hq[]``.
-    """
-    rematch = re.match
-    pat = re.compile(r"^{}\d*$".format(x))
-    return lambda s: rematch(pat, s)
-
-def getname(tree, accept_attr=True):
-    """The cousin of ``isx``.
-
-    From the same types of trees, extract the name as str.
-
-    If no match on ``tree``, return ``None``.
-    """
-    if type(tree) is Name:
-        return tree.id
-    if type(tree) is Captured:
-        return tree.name
-    if accept_attr and type(tree) is Attribute:
-        return tree.attr
-    return None
-
 def isec(tree, known_ecs):
     """Test whether tree is a call to a function known to be an escape continuation.
 
@@ -77,7 +36,7 @@ def isec(tree, known_ecs):
 
     **CAUTION**: Only bare-name references are supported.
     """
-    return type(tree) is Call and type(tree.func) is Name and tree.func.id in known_ecs
+    return type(tree) is Call and getname(tree.func, accept_attr=False) in known_ecs
 
 def detect_callec(tree):
     """Collect names of escape continuations from call_ec invocations in tree.
@@ -97,65 +56,85 @@ def detect_callec(tree):
             ...
         ...
         result = call_ec(g)  # <-- but this is here; g could be in another module
+
+    Additionally, the literal names `ec`, `brk`, `throw` are always interpreted
+    as invoking an escape continuation (whether they actually do or not).
+    So if you need the third pattern above, use **exactly** the name `ec`
+    for the escape continuation parameter, and it will work.
+
+    (The name `brk` covers the use of `unpythonic.fploop.breakably_looped`,
+    and `throw` covers the use of `unpythonic.ec.throw`.)
     """
-    # literal function names that are always interpreted as an ec.
-    # "brk" is needed to combo with unpythonic.fploop.breakably_looped.
-    fallbacks = ["ec", "brk"]
-    iscallec = partial(isx, make_isxpred("call_ec"))
-    @Walker
-    def detect(tree, *, collect, **kw):
-        # TODO: add support for general use of call_ec as a function (difficult)
-        if type(tree) in (FunctionDef, AsyncFunctionDef) and any(iscallec(deco) for deco in tree.decorator_list):
-            fdef = tree
-            collect(fdef.args.args[0].arg)  # FunctionDef.arguments.(list of arg objects).arg
-        elif is_decorated_lambda(tree, mode="any"):
-            decorator_list, thelambda = destructure_decorated_lambda(tree)
-            if any(iscallec(decocall.func) for decocall in decorator_list):
-                collect(thelambda.args.args[0].arg)  # we assume it's the first arg, as that's what call_ec expects.
-        return tree
-    return fallbacks + detect.collect(tree)
+    fallbacks = ["ec", "brk", "throw"]
+    iscallec = partial(isx, x="call_ec")
+    def detect(tree):
+        class Detector(ASTVisitor):
+            def examine(self, tree):
+                # TODO: add support for general use of call_ec as a function (difficult)
+                if type(tree) in (FunctionDef, AsyncFunctionDef) and any(iscallec(deco) for deco in tree.decorator_list):
+                    # TODO: Python 3.8+: handle the case where the ec is a positional-only arg?
+                    fdef = tree
+                    self.collect(fdef.args.args[0].arg)  # FunctionDef.arguments.(list of arg objects).arg
+                elif is_decorated_lambda(tree, mode="any"):
+                    # TODO: should we recurse selectively in this case?
+                    decorator_list, thelambda = destructure_decorated_lambda(tree)
+                    if any(iscallec(decocall.func) for decocall in decorator_list):
+                        self.collect(thelambda.args.args[0].arg)  # we assume it's the first arg, as that's what call_ec expects.
+                self.generic_visit(tree)
+        d = Detector()
+        d.visit(tree)
+        return d.collected
+    return fallbacks + detect(tree)
 
-@Walker
-def detect_lambda(tree, *, collect, stop, **kw):
-    """Find lambdas in tree. Helper for block macros.
+def detect_lambda(tree):
+    """Find lambdas in tree. Helper for two-pass block macros.
 
-    Run ``detect_lambda.collect(tree)`` in the first pass, before allowing any
-    nested macros to expand. (Those may generate more lambdas that your block
-    macro is not interested in).
+    A two-pass block macro first performs some processing outside-in, then calls
+    `expander.visit_recursively(tree)` to make any nested macro invocations expand,
+    and then performs some processing inside-out.
 
-    The return value from ``.collect`` is a ``list``of ``id(lam)``, where ``lam``
-    is a Lambda node that appears in ``tree``. This list is suitable as
-    ``userlambdas`` for the TCO macros.
+    Run ``detect_lambda(tree)`` in the outside-in pass, before calling
+    `expander.visit_recursively(tree)`, because nested macro invocations
+    may generate more lambdas that your block macro is not interested in.
+
+    The return value is a ``list``of ``id(lam)``, where ``lam`` is a Lambda node
+    that appears in ``tree``. This list is suitable as ``userlambdas`` for the
+    TCO macros.
 
     This ignores any "lambda e: ..." added by an already expanded ``do[]``,
     to allow other block macros to better work together with ``with multilambda``
-    (which expands in the first pass, to eliminate semantic surprises).
+    (which expands outside-in, to eliminate semantic surprises).
     """
-    if isdo(tree):
-        stop()
-        thebody = ExpandedDoView(tree).body
-        for thelambda in thebody:  # lambda e: ...
-            detect_lambda.collect(thelambda.body)
-    if type(tree) is Lambda:
-        collect(id(tree))
-    return tree
+    class LambdaDetector(ASTVisitor):
+        def examine(self, tree):
+            if isdo(tree, expanded=True):
+                thebody = ExpandedDoView(tree).body
+                for thelambda in thebody:  # lambda e: ...
+                    self.visit(thelambda.body)
+                return  # don't recurse generically, we already did it selectively
+            if type(tree) is Lambda:
+                self.collect(id(tree))
+            self.generic_visit(tree)
+    d = LambdaDetector()
+    d.visit(tree)
+    return d.collected
 
 def is_decorator(tree, fname):
     """Test tree whether it is the decorator ``fname``.
 
     ``fname`` may be ``str`` or a predicate, see ``isx``.
 
-    References of the forms ``f``, ``foo.f`` and ``hq[f]`` are supported.
+    References of the forms ``f``, ``foo.f`` and ``q[h[f]]`` are supported.
 
      We detect:
 
-        - ``Name``, ``Attribute`` or ``Captured`` matching the given ``fname``
-          (non-parametric decorator), and
+        - ``Name``, ``Attribute``, a `mcpyrate.core.Done`, or a `mcpyrate`
+          hygienic capture matching the given ``fname`` (non-parametric decorator), and
 
         - ``Call`` whose ``.func`` matches the above rule (parametric decorator).
     """
-    return isx(tree, fname) or \
-           (type(tree) is Call and isx(tree.func, fname))
+    return ((isx(tree, fname)) or
+            (type(tree) is Call and isx(tree.func, fname)))
 
 def is_lambda_decorator(tree, fname=None):
     """Test tree whether it decorates a lambda with ``fname``.
@@ -174,8 +153,8 @@ def is_lambda_decorator(tree, fname=None):
         trampolined(arg)                    # --> non-parametric decorator
         looped_over(range(10), acc=0)(arg)  # --> parametric decorator
     """
-    return (type(tree) is Call and len(tree.args) == 1) and \
-           (fname is None or is_decorator(tree.func, fname))
+    return ((type(tree) is Call and len(tree.args) == 1) and
+            (fname is None or is_decorator(tree.func, fname)))
 
 def is_decorated_lambda(tree, mode):
     """Detect a tree of the form f(g(h(lambda ...: ...)))
@@ -194,7 +173,7 @@ def is_decorated_lambda(tree, mode):
     assert mode in ("known", "any")
     if mode == "known":
         detectors = [partial(is_lambda_decorator, fname=x) for x in all_decorators]
-    else: # mode == "any":
+    else:  # mode == "any":
         detectors = [is_lambda_decorator]
 
     def detect(tree):
@@ -220,7 +199,7 @@ def destructure_decorated_lambda(tree):
             return get(tree.args[0], lst + [tree])
         elif type(tree) is Lambda:
             return lst, tree
-        assert False, "Expected a chain of Call nodes terminating in a Lambda node"
+        raise SyntaxError("Expected a chain of Call nodes terminating in a Lambda node")  # pragma: no cover
     return get(tree, [])
 
 def has_tco(tree, userlambdas=[]):
@@ -281,31 +260,37 @@ def sort_lambda_decorators(tree):
         for k, (pri, fname) in enumerate(decorator_registry):
             if is_lambda_decorator(tree, fname):
                 return k
-        x = getname(tree.func) if type(tree) is Call else "<unknown>"
-        assert False, "Only registered decorators can be auto-sorted, '{:s}' is not; see unpythonic.regutil".format(x)
+        # Only happens if called for a `tree` containing unknown (not registered) decorators.
+        x = getname(tree.func) if type(tree) is Call else tree  # pragma: no cover
+        raise SyntaxError(f"Only registered decorators can be auto-sorted, {repr(x)} is not; see unpythonic.regutil")  # pragma: no cover
 
-    @Walker
-    def fixit(tree, *, stop, **kw):
-        # we can robustly sort only decorators for which we know the correct ordering.
-        if is_decorated_lambda(tree, mode="known"):
-            decorator_list, thelambda = destructure_decorated_lambda(tree)
-            # We can just swap the func attributes of the nodes.
-            ordered_decorator_list = sorted(decorator_list, key=prioritize)
-            ordered_funcs = [x.func for x in ordered_decorator_list]
-            for thecall, newfunc in zip(decorator_list, ordered_funcs):
-                thecall.func = newfunc
-            # don't recurse on the tail of the "decorator list" (call chain),
-            # but recurse into the lambda body.
-            stop()
-            fixit.recurse(thelambda.body)
-        return tree
-    return fixit.recurse(tree)
+    class FixIt(ASTTransformer):
+        def transform(self, tree):
+            if is_captured_value(tree):
+                return tree  # don't recurse!
+            # we can robustly sort only decorators for which we know the correct ordering.
+            if is_decorated_lambda(tree, mode="known"):
+                decorator_list, thelambda = destructure_decorated_lambda(tree)
+                # We can just swap the func attributes of the nodes.
+                ordered_decorator_list = sorted(decorator_list, key=prioritize)
+                ordered_funcs = [x.func for x in ordered_decorator_list]
+                for thecall, newfunc in zip(decorator_list, ordered_funcs):
+                    thecall.func = newfunc
+                # don't recurse on the tail of the "decorator list" (call chain),
+                # but recurse into the lambda body.
+                thelambda.body = self.visit(thelambda.body)
+                return tree
+            return self.generic_visit(tree)
+    return FixIt().visit(tree)
 
 # TODO: should we just sort the decorators here, like we do for lambdas?
-# (The current solution is less magic, but less uniform.)
+# (The current solution is less magic, but also less uniform.)
 def suggest_decorator_index(deco_name, decorator_list):
     """Suggest insertion index for decorator deco_name in given decorator_list.
 
+    ``decorator_list`` is the eponymous attribute of a ``FunctionDef``
+    or ``AsyncFunctionDef`` AST node.
+
     The return value ``k`` is intended to be used like this::
 
         if k is not None:
@@ -318,7 +303,7 @@ def suggest_decorator_index(deco_name, decorator_list):
     is ``None``.
     """
     if deco_name not in all_decorators:
-        return None  # unknown decorator, don't know where it should go
+        return None  # unknown decorator, don't know where it should go  # pragma: no cover
     names = [getname(tree) for tree in decorator_list]
     pri_by_name = {dname: pri for pri, dname in decorator_registry}
 
@@ -345,11 +330,12 @@ def suggest_decorator_index(deco_name, decorator_list):
         return 0
     if targetpri > knownpris[-1]:
         return len(decorator_list)
+    assert knownpris[0] <= targetpri <= knownpris[-1]
     for pri, dname in zip(knownpris, knownnames):
-        if targetpri >= pri:
+        if pri >= targetpri:
             break
     else:
-        return None
+        assert False  # pragma: no cover
     return names.index(dname)
 
 def eliminate_ifones(body):
@@ -368,15 +354,14 @@ def eliminate_ifones(body):
     """
     def isifone(tree):
         if type(tree) is If:
-            if type(tree.test) is Num:
-                if tree.test.n == 1:
+            try:
+                value = getconstant(tree.test)
+            except TypeError:
+                pass
+            else:
+                if value in (1, True):
                     return "then"
-                elif tree.test.n == 0:
-                    return "else"
-            elif type(tree.test) is NameConstant:
-                if tree.test.value is True:
-                    return "then"
-                elif tree.test.value in (False, None):
+                elif value in (0, False, None):
                     return "else"
         return False
 
@@ -385,7 +370,7 @@ def eliminate_ifones(body):
         if t:
             branch = tree.body if t == "then" else tree.orelse
             return branch
-        return [tree]
+        return tree
 
     return transform_statements(optimize, body)
 
@@ -398,11 +383,11 @@ def transform_statements(f, body):
     the transformation ``f``.
 
     ``f`` is a one-argument function that takes an AST representing a single
-    statement, and returns a ``list`` of ASTs representing statements.
-    If the output is such a ``list``, it will be spliced to replace the
-    input statement. This allows ``f`` to drop a statement (1->0) or to
-    replace one statement with several (1->n), beside making one-to-one (1->1)
-    transformations.
+    statement, and that **must** return a ``list`` of ASTs representing statements.
+
+    The output ``list`` will be spliced to replace the input statement. This
+    allows ``f`` to drop a statement (1->0) or to replace one statement with
+    several (1->n), beside making one-to-one (1->1) transformations.
 
     (Transformations requiring n input statements are currently not supported.)
 
@@ -413,108 +398,46 @@ def transform_statements(f, body):
     original lists inside the ASTs containing the statements are in-place
     replaced with the transformed ones.
 
-    The return value is the transformed ``body``. It is always a ``list`` of ASTs.
+    The return value is the transformed ``body``.
     """
-    def rec(tree):
-        # TODO: brittle, may need changes as the Python AST evolves. Better ways to do this?
-        if type(tree) in (FunctionDef, AsyncFunctionDef, ClassDef, With, AsyncWith):
-            tree.body = rec(tree.body)
-        elif type(tree) in (If, For, While, AsyncFor):
-            tree.body = rec(tree.body)
-            tree.orelse = rec(tree.orelse)
-        elif type(tree) is Try:
-            tree.body = rec(tree.body)
-            tree.orelse = rec(tree.orelse)
-            tree.finalbody = rec(tree.finalbody)
-            for handler in tree.handlers:
-                handler.body = rec(handler.body)
-        elif type(tree) is list:  # multiple-statement body in AST
-            return [output_stmt for input_stmt in tree for output_stmt in rec(input_stmt)]
-        return f(tree)  # a single statement
-    return rec(body)
+    class StatementTransformer(ASTTransformer):
+        def transform(self, tree):
+            tree = self.generic_visit(tree)
+            if isinstance(tree, stmt):
+                return f(tree)
+            return tree
+    return StatementTransformer().visit(body)
 
-def splice(tree, rep, tag):
-    """Splice in a tree into another tree.
+# --------------------------------------------------------------------------------
+# AST markers.
 
-    Walk ``tree``, replacing the first occurrence of a ``Name(id=tag)`` with
-    the tree ``rep``.
+class UnpythonicASTMarker(ASTMarker):
+    """Base class for all AST markers used by `unpythonic`."""
+class UnpythonicExpandedMacroMarker(UnpythonicASTMarker):
+    """AST marker base class for expanded `unpythonic.syntax` macros."""
 
-    This is convenient for first building a skeleton with a marker such as
-    ``q[name["_here_"]]``, and then splicing in ``rep`` later. See ``forall``
-    and ``envify`` for usage examples.
-    """
-    @Walker
-    def doit(tree, *, stop, **kw):
-        if type(tree) is Name and tree.id == tag:
-            stop()
-            return rep
-        return tree
-    return doit.recurse(tree)
-
-def wrapwith(item, body, locref=None):
-    """Wrap ``body`` with a single-item ``with`` block, using ``item``.
-
-    ``item`` must be an ``ast.withitem``.
-
-    ``locref`` is an optional AST node to copy source location info from.
-    If not supplied, ``body[0]`` is used.
-
-    Syntax transformer. Returns the wrapped body.
-    """
-    locref = locref or body[0]
-    wrapped = With(items=[withitem(context_expr=item, optional_vars=None)],
-                   body=body,
-                   lineno=locref.lineno, col_offset=locref.col_offset)
-    return [wrapped]
-
-def ismarker(typename, tree):
-    """Return whether tree is a specific AST marker. Used by block macros.
-
-    That is, whether ``tree`` is a ``with`` block with a single context manager,
-    which is represented by a ``Name`` whose ``id`` matches the given ``typename``.
-
-    Example. If ``tree`` is the AST for the following code::
-
-        with ContinuationsMarker:
-            ...
-
-    then ``ismarker("ContinuationsMarker", tree)`` returns ``True``.
-    """
-    if type(tree) is not With or len(tree.items) != 1:
-        return False
-    ctxmanager = tree.items[0].context_expr
-    return type(ctxmanager) is Name and ctxmanager.id == typename
-
-# We use a custom metaclass to make __enter__ and __exit__ callable on the class
-# instead of requiring an instance.
-#
-# Note ``thing.dostuff(...)`` means ``Thing.dostuff(thing, ...)``; the method
-# is looked up *on the class* of the instance ``thing``, not on the instance
-# itself. Hence, to make method lookup succeed when we have no instance, the
-# method should be defined on the class of the class, i.e. *on the metaclass*.
-# https://stackoverflow.com/questions/20247841/using-delitem-with-a-class-object-rather-than-an-instance-in-python
-class ASTMarker(type):
-    """Metaclass for AST markers used by block macros.
-
-    This can be used by block macros to tell other block macros that a section
-    of the AST is an already-expanded block of a given kind (so that others can
-    tune their processing or skip it, as appropriate). At run time a marker
-    does nothing.
-
-    Usage::
-
-        with SomeMarker:
-            ... # expanded code goes here
-
-    We provide a custom metaclass so that there is no need to instantiate
-    ``SomeMarker``; suitable no-op ``__enter__`` and ``__exit__`` methods
-    are defined on the metaclass, so e.g. ``SomeMarker.__enter__`` is valid.
-    """
-    def __enter__(cls):
-        pass
-    def __exit__(cls, exctype, excvalue, traceback):
-        pass
-
-class ContinuationsMarker(metaclass=ASTMarker):
+class ExpandedContinuationsMarker(UnpythonicExpandedMacroMarker):
     """AST marker for an expanded "with continuations" block."""
-    pass
+
+class ExpandedAutorefMarker(UnpythonicExpandedMacroMarker):
+    """AST marker for an expanded "with autoref[o]" block."""
+    def __init__(self, body, varname):
+        super().__init__(body)
+        self.varname = varname
+        self._fields += ["varname"]
+
+# The point of having these two functions is:
+#   - `__init__` must explicitly enable the hook, thus making its existence
+#     obvious, since the entry-point source file for the macro layer has an
+#     obvious function call, instead of having the hook secretly registered
+#     by an innocuous-looking utility module.
+#
+#   - We could register `partial(delete_markers, cls=UnpythonicASTMarker)`,
+#     but then we would be unable to `remove_postprocessor` it later, because
+#     the function object itself is the key used for unregistering. Currently
+#     we don't need to do that, but it's nice to have the possibility.
+def register_postprocessor_hook():
+    """Set up global postprocessor hook for `mcpyrate` to nuke `unpythonic`'s AST markers from the final tree."""
+    add_postprocessor(_delete_unpythonic_ast_markers)
+def _delete_unpythonic_ast_markers(tree):
+    return delete_markers(tree, cls=UnpythonicASTMarker)
